#!/usr/bin/env python

__copyright__ = "(C) 2006-2007 Ola Skavhaug and Simula Research Laboratory"
__license__   = "GNU LGPL Version 2.1."
__cite__      = """Ola Skavhaug, Viper Visualization Software, 
http://www.fenics.org/viper/"""
__version__ = "0.2.0"

r"""
Viper

A simple mesh plotter and run--time visualization module for plotting and
saving simulation data. The class C{Viper} can visualize solutions given as
numpy arrays, and meshes that provide the two methods C{cells()} and
C{coordinates()}. These methods should return numpy arrays specifying the
node-element ordering and coordinates of the nodes, respectively.

Copyright: %s
Licence: %s
Citation: %s

""" % (__copyright__, __license__, __cite__)


import vtk
from vtk.util.colors import tomato
import numpy
import os
import math

_viper = None

lutdir = os.path.join(os.path.split(__file__)[0], "data")

class Viper(object):
    """Simple real-time vtk plotter."""

    def __init__(self, mesh=None, x=None, vmin=None, vmax=None, lutfile="gauss_120.lut", wireframe=False, displacement=None, celltype=None): 
        """Constructor takes a grid and a field on that grid. Based on this
        input, a vtk grid is created, a plot render window is created, and the
        field x is plotted."""

        self._update = self.update

        self.initcommon()
        if mesh is not None:
            self.mesh = mesh
            self.lutfile = lutfile
            self.x = x
            self.vtkgrid = self.make_vtk_grid(cell_type=celltype)
            self._vtkgrid = self.vtkgrid
            
            if vmin is None:
                vmin = x.min()
            if vmax is None:
                vmax = x.max()
            
            self.displacement = self._resolve_displacement(displacement)
            
            if len(x.shape) == 1:
                (self.iren, self.renWin, self.ren) = self.simple_plotter(self.vtkgrid, vmin, vmax, wireframe)
            else:
                (self.iren, self.renWin, self.ren) = self.vector_plotter(self.vtkgrid, self.mesh.coordinates(), x, vmin, vmax, wireframe)
            self._update(x=x)
            self.iren.Initialize()

    def init_from_file(self, filename):
        """Construct a scalar field over a unstructured grid from vtk file."""
        reader = vtk.vtkUnstructuredGridReader()
        reader.SetFileName(filename)
        reader.Update()
        self.vtkgrid = reader.GetOutput()
        self._vtkgrid = self.vtkgrid
        pd = self.vtkgrid.GetPointData()
        scalars = pd.GetScalars()
        n = scalars.GetNumberOfTuples()
        x = numpy.zeros(n)
        scalars.ExportToVoidPointer(x)
        self.x = x
        self.displacement = None

        self.lutfile = "gauss_120.lut"
        self.refs = []
        self.basename = None

        self.elevator = 0
        self.elevator_sign = 1
        self.set_camera_movement()
        self.darr = vtk.vtkDoubleArray()

        (self.iren, self.renWin, self.ren) = self.simple_plotter(self.vtkgrid, min(x), max(x), False)
        self._update(self.x)


    def _resolve_displacement(self, d):
        if d is None:
            return None

        if len(d.shape) > 1 and d.shape[1] > 1:
            d = self.vec3d(d)
            self.vtkgrid = self.warp_vector(d)
            return d
        else:
            self.vtkgrid = self.warp_scalar(d)
            return d
 
    def initcommon(self):
        self.filters = []
        self.refs = []
        self.basename = None
        self.displacement = None
        self.outline = False
        self.vertex_plot = True
        self.plottype = None
        self.mode = None
        self.iren = None
        self.ren = None
        self.renWin = None

        # paramaters to control camera movement, default is no movement
        self.elevator = 0
        self.elevator_sign = 1
        self.set_camera_movement()
        self.darr = vtk.vtkDoubleArray()

    def warp_scalar(self, d, minmax=1.0):
        if len(d.shape) > 1 and d.shape[1] > 1:
            raise ValueError, "Wrong shape in scalar displacement field"
        self.disp_arr = vtk.vtkDoubleArray()
        self.disp_arr.SetNumberOfComponents(1)
        self.disp_arr.SetVoidArray(d, d.shape[0], 1)
        self._vtkgrid.GetPointData().SetScalars(self.disp_arr)
        self.warp = vtk.vtkWarpScalar()
        self.warp.SetInput(self._vtkgrid)
        self.warp.SetScaleFactor(1.0/minmax)
        self.warpdata = self.warp.GetOutput()
        return self.warpdata

    def warp_vector(self, d):
        if not d.shape == (len(self.x), 3):
            raise ValueError, "Wrong shape in vector displacement field (must be 3d)"

        self.disp_arr = vtk.vtkDoubleArray()
        self.disp_arr.SetNumberOfComponents(3)
        self.disp_arr.SetVoidArray(d, d.shape[0]*d.shape[1], 1)
        self._vtkgrid.GetPointData().SetVectors(self.disp_arr)

        self.warp = vtk.vtkWarpVector()
        self.warp.SetInput(self._vtkgrid)
        self.warp.SetScaleFactor(1.0)
        self.warpdata = self.warp.GetOutput()
        return self.warpdata

    def set_contour(self, nlevels=5):
        self.contour = vtk.vtkContourFilter()

    def init_writer(self, filebasename="simulation"):
        """Initialize the simple vtk file writer for storing unstructured grids and fields to
        file."""

        writepath = os.path.dirname(filebasename)
        if writepath and not os.path.isdir(writepath):
            os.makedirs(writepath)

        self.writer = vtk.vtkUnstructuredGridWriter()
        self.writer.SetFileTypeToBinary()
        self.basename = filebasename
        self.filecounter = -1
        self.writer.SetInput(self.vtkgrid)
        
        w2if = vtk.vtkWindowToImageFilter()
        w2if.SetInput (self.renWin)
        self.pngwriter = vtk.vtkPNGWriter()
        self.pngfilecounter = -1
        self.pngwriter.SetInput(w2if.GetOutput())
        self.w2if = w2if

        self.rawfilecounter = -1

    def write_vtk(self, filename=None):
        """Write data to file. This works because the attached field self.x is
        a pointer to the data that is being computed."""
        if not self.basename: self.init_writer()
        self.filecounter += 1
        if filename is None:
            filename = "%s%.4d.vtk" % (self.basename, self.filecounter)
        self._update()
        self.writer.SetFileName(filename)
        self.writer.Write()

    def write_png(self, filename=None):
        """Write a simulation frame to file in png format.""" 
        if not self.basename: self.init_writer()
        self.pngfilecounter += 1
        if filename is None:
            filename = "%s%.4d.png" % (self.basename, self.pngfilecounter)
        self._update()
        self.pngwriter.SetFileName(filename)
        self.w2if.Modified()
        self.pngwriter.Write()

    def write_raw(self, sol=None, dofmap=None):
        """Write a simulation frame to file, i.e. dump the numpy array to file
        in binary format.""" 
        if not self.basename: self.init_writer()
        #self._update(sol)
        self.rawfilecounter += 1

        if not dofmap is None:
            self.x[dofmap].dump("%s%.4d.raw" % (self.basename, self.rawfilecounter))
        else:
            self.x.dump("%s%.4d.raw" % (self.basename, self.rawfilecounter))

    def interactive(self):
        """Hand the control over to the render window."""
        self.iren.Start()

    def azimuth(self, angle):
        self.ren.GetActiveCamera().Azimuth(angle)

    def elevate(self, x):
        self.ren.GetActiveCamera().Elevation(x)

    def dolly(self, x):
        self.ren.GetActiveCamera().Dolly(x)

    def set_viewangle(self, i):
        self.ren.GetActiveCamera().SetViewAngle(i)

    def set_camera_movement(self, a=0, e=0, m=60):
        self.azmiuth_incr = a;   # 1 (degree) is a good value
        self.elevator_incr = e;  # 1 (degree) is a good value
        self.max_elevation = m;
        
    def _make_lut(self, autorange=(0,0)):
        lut = vtk.vtkLookupTable()
        lutfile = os.path.join(lutdir, self.lutfile)
        if os.path.isfile(self.lutfile):
            lutfile = self.lutfile
        if os.path.isfile(lutfile):
            vals = [x.split() for x in open(lutfile,'r').readlines()[1:]]
            lut.SetNumberOfColors(len(vals))
            lut.Build()
            for i in range(len(vals)): 
                if len(vals[i]) == 4:
                    lut.SetTableValue(i, *[float(x) for x in vals[i]]),
        else:
            rmin, rmax = autorange
            ncols = int(rmax-rmin)+1
            lut.SetNumberOfColors(ncols)
            lut.Build()
        return lut


    def make_vtk_grid(self, mesh=None, cell_type=None):
        """Based on a mesh with the methods cells() and coordinates(), construct a vtk grid."""
        if mesh is None: mesh = self.mesh
        cell_dim = mesh.cells().shape[1]
        if cell_dim == 2:
            celltype=3
        elif cell_dim == 3:
            celltype = 5
        elif cell_dim == 4:
            celltype = 10
        elif cell_type == "quads":
            celltype = 9

        cellsize = cell_dim + 1

        if cell_type == "quads":
            cellsize = 4

        nl = mesh.cells().copy()
        _cl = mesh.coordinates()
        cl = numpy.zeros((_cl.shape[0], 3), dtype='d')

        if _cl.shape[1] == 2: 
            cl[:,:2] = _cl.copy()
        else: 
            cl = _cl.copy()

        inttype = nl.dtype
            
        # Create a cell array, stored in Python.numpy
        ncells = len(nl)
        cells = numpy.zeros((ncells, cellsize),dtype=inttype)
            
        cells[:,1:cellsize] = nl
        cells[:,0] = cellsize - 1 
            
        a = numpy.ravel(cells)
        ita = vtk.vtkIdTypeArray()
        ita.SetVoidArray(a, len(a), 1)
        ca = vtk.vtkCellArray()
        ca.SetCells(ncells,ita)
        
            
        # Create some points
        npoints = len(cl)
        pts = numpy.ravel(cl)
            
        pa = vtk.vtkDoubleArray()
        pa.SetNumberOfComponents(3)
        pa.SetVoidArray(pts, npoints*3, 1)
        
        self.refs = []
        self.refs.append(a)
        self.refs.append(pts)
            
        v = vtk.vtkPoints()
        v.SetNumberOfPoints(npoints)
        v.SetData(pa)
            
        # Create an unstructured grid.
        us = vtk.vtkUnstructuredGrid()
        us.SetPoints(v)
        us.SetCells(celltype, ca)
        
        return us

    def vec3d(self, x):
        nsd = 3
        (n,m) = x.shape
        if m < nsd:
            x_old = x
            x = numpy.zeros((n, nsd))
            x[:,:m] = x_old
        return x

    
    def update(self, x=None):
        """Update plot data."""

        if self.displacement is not None:
            self.disp_arr.Modified()

        if not x is None:
            self.x = x
            if len(self.x.shape) == 1:
                self.darr.SetVoidArray(self.x, len(self.x), 1)
            else:
                nsd = 3
                self.darr.SetNumberOfComponents(nsd)
                self.x = self.vec3d(self.x)
                self.x.shape=(nsd*len(self.x),)
                self.darr.SetVoidArray(self.x, len(self.x), 1)

        self.darr.Modified()
        self.ren.ResetCameraClippingRange()

        if self.vertex_plot:
            if len(self.x.shape) == 1:
                self.vtkgrid.GetPointData().SetScalars(self.darr)
            else:
                self.vtkgrid.GetPointData().SetVectors(self.darr)
        else:
            if len(self.x.shape) == 1:
                self.vtkgrid.GetCellData().SetScalars(self.darr)

        self.renWin.Render()

        self.azimuth(self.azmiuth_incr)
        # assumes that initial elevation is zero. If not the sum of
        # elevate steps might exceed 90 degrees, which is not good.
        # Fix: use something like GetActiveCamera().getElevation()
        self.elevator += self.elevator_incr;
        step = 1.5*self.elevator_incr*math.cos(self.elevator*3.1416/180.0)
        self.elevate(step)

    def _add_arrow(self, vectors):
        arrow = vtk.vtkArrowSource()
        arrow.SetTipRadius(0.15)
        arrow.SetTipLength(0.15)
        arrow.SetShaftRadius(0.05)
        vectors.SetSource(arrow.GetOutput())
        vectors.SetScaleFactor(0.01)
        return vectors

    def set_vector_scale(self, scale_factor):
        """Set the vector scale factor."""
        self.vectors.SetScaleFactor(scale_factor)
        self.vectors.Modified()
        self.renWin.Render()

    def _vectors_from_numpy(self, coords, directions):
        vectors = vtk.vtkGlyph3D()
        vectors.SetColorModeToColorByVector ()
        vectors.SetScaleModeToScaleByVector()
        coordinates = coords
        (n,m) = coordinates.shape
        coords_3d = numpy.zeros((n, 3), dtype='d')
        coords_3d[:,:m] = coordinates
        directions_3d = numpy.zeros((n, 3), dtype='d')
        if directions.shape[1] > m: # If direcrions is 3d already
            m = directions.shape[1]
        directions_3d[:,:m] = directions

        coordinates = coords_3d
        directions = directions_3d
        m = 3
                                
        vtkpoints = vtk.vtkDoubleArray()
        vtkpoints.SetNumberOfComponents(m)
        vtkpoints.SetVoidArray(coordinates, n*m, m)
        _points = vtk.vtkPoints()
        _points.SetData(vtkpoints)

        vtkdirections = vtk.vtkDoubleArray()
        vtkdirections.SetNumberOfComponents(m)
        vtkdirections.SetVoidArray(directions, n*m, m)

        data = vtk.vtkPolyData()
        data.SetPoints(_points)
        data.GetPointData().SetVectors(vtkdirections)
        vectors.SetInput(data)
        vectors.SetVectorModeToUseVector()
        self.coordinates = coordinates
        self.directions = directions
        self.vtkpoints = vtkpoints
        self.vtkdirections = vtkdirections
        self.data = data
        return self._add_arrow(vectors)

    def _construct_interactor(self):
        style = vtk.vtkInteractorStyleSwitch()
        style.SetCurrentStyleToTrackballCamera()
        iren = vtk.vtkRenderWindowInteractor()
        iren.SetInteractorStyle(style)
        return iren

    def _construct_renderer(self):
        ren = vtk.vtkRenderer()
        renWin = vtk.vtkRenderWindow()
        renWin.SetWindowName("FEniCS Viper")
        renWin.AddRenderer(ren)
        renWin.SetSize(600, 400)
        return ren, renWin

    def vector_plotter(self, data, coords, x, vmin, vmax, wireframe=False):
        import operator
        vectors = self._vectors_from_numpy(coords, x)
        # Resolve scale factor:
        norms = [numpy.linalg.norm(v) for v in x]
        vmin, vmax = min(norms), max(norms)
        nnz, nsd = coords.shape
        a = self.simple_outline()[0].GetBounds()
        D = math.sqrt(reduce(operator.add,[(a[i]-a[i-1])**2 for i in xrange(1,6,2)]))
        sf = 2.0/vmax*D/(nnz**(1.0/nsd))
        vectors.SetScaleFactor(sf)

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInput(vectors.GetOutput())
        vactor = vtk.vtkActor()
        vactor.SetMapper(mapper)
        vactor.AddPosition(0, 0, 0)
        self.vectors = vectors

        ren, renWin = self._construct_renderer()
        iren        = self._construct_interactor()
        iren.SetRenderWindow(renWin)

        ren.SetBackground(1, 1, 1)
        ren.AddActor(vactor)
        ren.ResetCamera()
        ren.GetActiveCamera().Azimuth(00)
        ren.GetActiveCamera().Elevation(0)
        ren.GetActiveCamera().Dolly(1.5)
        ren.ResetCameraClippingRange()
        iren.AddObserver("KeyPressEvent", self.key_press_methods)
        return iren, renWin, ren


#    def simple_vector_plotter(self, data, x, vmin, vmax, wireframe=False):
#        """Construct a simple vector vtk plotter for data."""
#        mapper = vtk.vtkPolyDataMapper()
#        mapper.SetInput(vectors.GetOutput())
#
#        darr = vtk.vtkDoubleArray()
#        nsd = 3
#        darr.SetNumberOfComponents(nsd)
#        x = self.vec3d(x)
#        x.shape=(nsd*len(x),)
#        darr.SetVoidArray(x, len(x), 1)
#        data.GetPointData().SetVectors(darr)
#        self.darr = darr
#
#        hh = vtk.vtkHedgeHog()
#        hh.SetInput(data)
#        self.hh = hh
#        mapper.SetInput(hh.GetOutput())
#        lut = self._make_lut(autorange=(vmin, vmax))
#        mapper.SetLookupTable(lut)
#        mapper.SetScalarRange(vmin, vmax)
#        vactor = vtk.vtkActor()
#        vactor.SetMapper(mapper)
#        vactor.AddPosition(0, 0, 0)
##        if wireframe:
##            actor.GetProperty().SetRepresentationToWireframe ()
#        self.refs.append(mapper)
#
#
#        self.appendData = vtk.vtkAppendPolyData()
#        self.smapper = vtk.vtkPolyDataMapper()
#        self.sactor = vtk.vtkActor()
#
#        ren, renWin = self._construct_renderer()
#        iren        = self._construct_interactor()
#        iren.SetRenderWindow(renWin)
#        ren.SetBackground(1, 1, 1)
#        ren.AddActor(self.sactor)
#        ren.AddActor(vactor)
#        ren.ResetCamera()
#        ren.GetActiveCamera().Azimuth(00)
#        ren.GetActiveCamera().Elevation(0)
#        ren.GetActiveCamera().Dolly(1.5)
#        ren.ResetCameraClippingRange()
#        return iren, renWin, ren

    def display_keybindings(self):
        print """\
Keybindings:

Viper specific:
* Keypress v: write data to file in vtk format
* Keypress i: write image to file in png format
* Keypress o: add outline 
* Keypress h: display this message

Inherited from vtk:
* Keypress j / Keypress t: toggle between joystick (position sensitive) and trackball (motion sensitive) styles. In joystick style, motion occurs continuously as long as a mouse button is pressed. In trackball style, motion occurs when the mouse button is pressed and the mouse pointer moves.

* Keypress c / Keypress a: toggle between camera and actor modes. In camera mode, mouse events affect the camera position and focal point. In actor mode, mouse events affect the actor that is under the mouse pointer.

* Button 1: rotate the camera around its focal point (if camera mode) or rotate the actor around its origin (if actor mode). The rotation is in the direction defined from the center of the renderer's viewport towards the mouse position. In joystick mode, the magnitude of the rotation is determined by the distance the mouse is from the center of the render window.

* Button 2: pan the camera (if camera mode) or translate the actor (if actor mode). In joystick mode, the direction of pan or translation is from the center of the viewport towards the mouse position. In trackball mode, the direction of motion is the direction the mouse moves. (Note: with 2-button mice, pan is defined as <Shift>-Button 1.)

* Button 3: zoom the camera (if camera mode) or scale the actor (if actor mode). Zoom in/increase scale if the mouse position is in the top half of the viewport; zoom out/decrease scale if the mouse position is in the bottom half. In joystick mode, the amount of zoom is controlled by the distance of the mouse pointer from the horizontal centerline of the window.

* Keypress 3: toggle the render window into and out of stereo mode. By default, red-blue stereo pairs are created. Some systems support Crystal Eyes LCD stereo glasses; you have to invoke SetStereoTypeToCrystalEyes() on the rendering window.

* Keypress e: exit the application.

* Keypress f: fly to the picked point

* Keypress p: perform a pick operation. The render window interactor has an internal instance of vtkCellPicker that it uses to pick.

* Keypress r: reset the camera view along the current view direction. Centers the actors and moves the camera so that all actors are visible.

* Keypress s: modify the representation of all actors so that they are surfaces.

* Keypress u: invoke the user-defined function. Typically, this keypress will bring up an interactor that you can type commands in.

* Keypress w: modify the representation of all actors so that they are wireframe.

"""

    def simple_outline(self):
        outline = vtk.vtkOutlineFilter()
        outline.SetInput(self.vtkgrid)
        map = vtk.vtkPolyDataMapper()
        map.SetInputConnection(outline.GetOutputPort())
        actor = vtk.vtkActor()
        actor.SetMapper(map)
        actor.GetProperty().SetColor(1, 0, 0)
        return map, actor

    def key_press_methods(self, obj, event):
        key = obj.GetKeyCode()
        if key == "o":
            if not self.outline:
                map, actor = self.simple_outline()
                self.ren.AddActor(actor)
                self.renWin.Render()
                if True: # TODO: Figure out a way to steer this
                    bounds = map.GetBounds()
                    print "Size of bounding box:"
                    print "dx = ", bounds[1] - bounds[0]
                    print "dy = ", bounds[3] - bounds[2]
                    print "dz = ", bounds[5] - bounds[4]
                self.outline = True
        elif key == "i":
            self.write_png()
        elif key == "v":
            self.write_vtk()
        elif key == "h":
            self.display_keybindings()

    def begin_interaction(self, obj, event):
        self.glyphActor.VisibilityOn()
        print "Picker activated"

    def enable_event(self, obj, event):
        self.glyphActor.VisibilityOn()
        print "Interaction is ON, press again to turn off."        

    def disable_event(self, obj, event):
        self.glyphActor.VisibilityOff()
        print "Interaction is OFF, press again to turn on."        

    def probe_data(self, obj, event):
        self.glyphActor.VisibilityOff()
        obj.GetPolyData(self.point)
        print obj.GetPosition()

    def user_method(self, obj, event):
        print "Please refrain form pressing 'u'"
        print self.pointWidget.GetPosition()
        self.add_sphere(self.pointWidget.GetPosition(), 0.2)
        self._update()

    def simple_plotter(self, data, vmin, vmax, wireframe=False):
        """Construct a simple vtk plotter for data."""
        lut = self._make_lut(autorange=(vmin, vmax))
        mapper = vtk.vtkDataSetMapper()
        mapper.SetInput(data)
        mapper.SetLookupTable(lut)
        mapper.SetScalarRange(vmin, vmax)
        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.AddPosition(0, 0, 0)
        if wireframe:
            actor.GetProperty().SetRepresentationToWireframe ()
        self.lut = lut
        self.mapper = mapper
        self.actor = actor

        # The plane widget is used probe the dataset.
        pointWidget = vtk.vtkPointWidget()
        pointWidget.SetInput(data)
        pointWidget.AllOff()
        pointWidget.PlaceWidget()
        point = vtk.vtkPolyData()
        pointWidget.GetPolyData(point)

        
        probe = vtk.vtkProbeFilter()
        probe.SetInput(point)
        probe.SetSource(data)
        
        # create glyph
        cone = vtk.vtkConeSource()
        cone.SetResolution(16)
        glyph = vtk.vtkGlyph3D()
        glyph.SetInput(probe.GetOutput())
        glyph.SetSource(cone.GetOutput())
        glyph.SetVectorModeToUseVector()
        glyph.SetScaleModeToDataScalingOff()
        glyph.SetScaleFactor(data.GetLength()*0.1)
        glyphMapper = vtk.vtkPolyDataMapper()
        glyphMapper.SetInput(glyph.GetOutput())
        glyphActor = vtk.vtkActor()
        glyphActor.SetMapper(glyphMapper)
        glyphActor.VisibilityOff()

        probe = vtk.vtkProbeFilter()
        probe.SetInput(point)
        probe.SetSource(data)

        self.cone = cone
        self.glyph = glyph
        self.probe = probe
        self.point = point
        self.glyphActor = glyphActor
        self.pointWidget = pointWidget

        self.appendData = vtk.vtkAppendPolyData()
        self.smapper = vtk.vtkPolyDataMapper()
        self.sactor = vtk.vtkActor()

        if self.ren is None:
            ren, renWin = self._construct_renderer()
            iren        = self._construct_interactor()
            iren.SetRenderWindow(renWin)

            iren.AddObserver("KeyPressEvent", self.key_press_methods)
            ren.SetBackground(1, 1, 1)
        else:
            ren = self.ren
            renWin = self.renWin
            iren = self.iren
            actors = ren.GetActors() 
            actors.InitTraversal()
            for i in xrange(actors.GetNumberOfItems()):
                ren.RemoveActor(actors.GetNextItem())

        ren.AddActor(self.sactor)
        ren.AddActor(actor)
        ren.AddActor(glyphActor)
        ren.ResetCamera()
        ren.GetActiveCamera().Azimuth(00)
        ren.GetActiveCamera().Elevation(0)
        ren.GetActiveCamera().Dolly(1.5)
        ren.ResetCameraClippingRange()

        if False:
            iren.AddObserver("UserEvent", self.user_method)
            pointWidget.SetInteractor(iren)
            pointWidget.AddObserver("EnableEvent", self.enable_event)
            pointWidget.AddObserver("DisableEvent", self.disable_event)
            pointWidget.AddObserver("StartInteractionEvent", self.begin_interaction)
            pointWidget.AddObserver("InteractionEvent", self.probe_data)

        return iren, renWin, ren
    
    def add_sphere(self, pt, rad):
        """Add a sphere at point pt, with radius rad."""
        sphere = vtk.vtkSphereSource()
        sphere.SetThetaResolution(8)
        sphere.SetPhiResolution(6)
        sphere.SetRadius(rad)
        sphere.SetCenter(*pt)

        self.appendData.AddInput(sphere.GetOutput())
        self.smapper.SetInput(self.appendData.GetOutput())
        self.sactor.SetMapper(self.smapper)
        self.sactor.GetProperty().SetColor(0, 0, 0)

    def add_stim_sites(self, stim):
        """Add a sphere and a label to each stimulation site."""
        i = 0
        for item in stim:
            # draw a sphere at the stim site:
            point = item[0]
            if (len(point)==2):
                point =  (point[0], point[1], 0.0)
            self.add_sphere(point, item[1])
            # draw a lable at the stim site:
            act = vtk.vtkCaptionActor2D();
            act.SetCaption(str(i)); i += 1
            act.SetAttachmentPoint(point)
            act.BorderOff()
            act.GetProperty().SetColor(1, 0, 0)
            self.ren.AddActor(act)

    def set_sphere_opacity(self, val):
        """Change the sphere opacity (range [0,1])."""
        if val < 0.0: val = 0.0
        elif val > 1.0: val = 1.0

        self.sactor.GetProperty().SetOpacity(val)

    def set_min_max(self, min, max):
        """Set min and max scalar range"""
        self.mapper.SetScalarRange(min, max)
        self._update()


def plot(mesh, data, *args, **kwargs):
    global _viper
    _viper = Viper(mesh, data, *args, **kwargs)
    return _viper

def update(data):
    global _viper
    if _viper != None:
        _viper.update(data)
        return _viper
    print "No plot object, cannot update"

def interactive():
    if _viper != None:
        _viper.interactive()
        return _viper
    print "No plot object, interaction not possible"

def save_plot(data, mesh, filename="plot.png"):
    return Viper(data, mesh, filename=filename, interactive=False)
